@@ -33,8 +33,9 @@ void clockadj_init(clockid_t clkid);
  * Set clock's frequency offset.
  * @param clkid A clock ID obtained using phc_open() or CLOCK_REALTIME.
  * @param freq  The frequency offset in parts per billion (ppb).
+ * @return      0 on success, -1 on failure
  */
-void clockadj_set_freq(clockid_t clkid, double freq);
+int clockadj_set_freq(clockid_t clkid, double freq);
 
 /**
  * Read clock's frequency offset.
@@ -47,15 +48,17 @@ double clockadj_get_freq(clockid_t clkid);
  * Set clock's phase offset.
  * @param clkid  A clock ID obtained using phc_open() or CLOCK_REALTIME.
  * @param offset The phase offset in nanoseconds.
+ * @return       0 on success, -1 on failure
  */
-void clockadj_set_phase(clockid_t clkid, long offset);
+int clockadj_set_phase(clockid_t clkid, long offset);
 
 /**
  * Step clock's time.
  * @param clkid A clock ID obtained using phc_open() or CLOCK_REALTIME.
  * @param step  The time step in nanoseconds.
+ * @return      0 on success, -1 on failure
  */
-void clockadj_step(clockid_t clkid, int64_t step);
+int clockadj_step(clockid_t clkid, int64_t step);
 
 /**
  * Read maximum frequency adjustment of the target clock.
@@ -63,6 +66,24 @@ void clockadj_step(clockid_t clkid, int64_t step);
  */
 int clockadj_max_freq(clockid_t clkid);
 
+/**
+ * Compare offset between two clocks
+ * @param clkid     A clock ID obtained using phc_open() or CLOCK_REALTIME
+ * @param sysclk    A clock ID obtained using phc_open() or CLOCK_REALTIME
+ * @param readings  Number of readings to try
+ * @param offset    On return, the nanoseconds offset between the clocks
+ * @param ts        On return, the time of sysclk in nanoseconds that was used
+ * @param delay     On return, the interval between two reads of sysclk
+ * @return Zero on success, or negative error code on failure.
+ *
+ * Compare the offset between two clocks in a similar manner as the
+ * PTP_SYS_OFFSET ioctls. Performs multiple reads of sysclk with a read of
+ * clkid between in order to calculate the time difference of sysclk minus
+ * clkid.
+ */
+int clockadj_compare(clockid_t clkid, clockid_t sysclk, int readings,
+		     int64_t *offset, uint64_t *ts, int64_t *delay);
+
 /**
  * Set the system clock to insert/delete leap second at midnight.
  * @param leap  +1 to insert leap second, -1 to delete leap second,
