@@ -37,6 +37,9 @@
 
 #define MAJOR_VERSION_MASK	0x0f
 
+/* Values for the transportSpecific field */
+#define TS_IEEE_8021AS (1<<4)
+
 /* Values for the messageType field */
 #define SYNC                  0x0
 #define DELAY_REQ             0x1
@@ -438,6 +441,11 @@ static inline Boolean msg_unicast(struct ptp_message *m)
 	return field_is_set(m, 0, UNICAST);
 }
 
+/**
+ * Work around HW assuming PTP message version 2.0
+ */
+extern uint8_t ptp_hdr_ver;
+
 /**
  * Work around buggy 802.1AS switches.
  */
