@@ -274,9 +274,8 @@ static int nsm_open(struct nsm *nsm, struct config *cfg)
 	STAILQ_FOREACH(iface, &cfg->interfaces, list) {
 		ifname = interface_name(iface);
 		memset(ts_label, 0, sizeof(ts_label));
-		rtnl_get_ts_device(ifname, ts_label);
-		interface_set_label(iface, ts_label);
-		interface_ensure_tslabel(iface);
+		if (!rtnl_get_ts_device(ifname, ts_label))
+			interface_set_label(iface, ts_label);
 		count++;
 	}
 	if (count != 1) {
@@ -395,7 +394,6 @@ static int nsm_request(struct nsm *nsm, char *target)
 	msg->header.correction         = -asymmetry;
 	msg->header.sourcePortIdentity = nsm->port_identity;
 	msg->header.sequenceId         = nsm->sequence_id++;
-	msg->header.control            = CTL_DELAY_REQ;
 	msg->header.logMessageInterval = 0x7f;
 
 	msg->address = dst;
