@@ -20,9 +20,12 @@
 
 #include <ctype.h>
 #include <errno.h>
+#include <glob.h>
 #include <libgen.h>
 #include <limits.h>
+#include <time.h>
 #include <linux/net_tstamp.h>
+#include <net/if.h>
 #include <signal.h>
 #include <spawn.h>
 #include <stdarg.h>
@@ -31,10 +34,12 @@
 #include <string.h>
 #include <sys/stat.h>
 #include <sys/types.h>
+#include <sys/utsname.h>
 #include <sys/wait.h>
 #include <unistd.h>
 
 #include "print.h"
+#include "rtnl.h"
 #include "sk.h"
 #include "util.h"
 #include "version.h"
@@ -42,6 +47,8 @@
 #define DEFAULT_RUNDIR "/var/run/timemaster"
 
 #define DEFAULT_FIRST_SHM_SEGMENT 0
+#define DEFAULT_RESTART_PROCESSES 1
+#define DEFAULT_USE_VCLOCKS -1
 
 #define DEFAULT_NTP_PROGRAM CHRONYD
 #define DEFAULT_NTP_MINPOLL 6
@@ -106,6 +113,8 @@ struct timemaster_config {
 	enum ntp_program ntp_program;
 	char *rundir;
 	int first_shm_segment;
+	int restart_processes;
+	int use_vclocks;
 	struct program_config chronyd;
 	struct program_config ntpd;
 	struct program_config phc2sys;
@@ -117,9 +126,18 @@ struct config_file {
 	char *content;
 };
 
+struct phc_vclocks {
+	int pclock_index;
+	int vclocks;
+};
+
 struct script {
+	struct phc_vclocks **vclocks;
 	struct config_file **configs;
 	char ***commands;
+	int **command_groups;
+	int restart_groups;
+	int no_restart_group;
 };
 
 static void free_parray(void **a)
@@ -383,6 +401,10 @@ static int parse_timemaster_settings(char **settings,
 			replace_string(value, &config->rundir);
 		} else if (!strcasecmp(name, "first_shm_segment")) {
 			r = parse_int(value, &config->first_shm_segment);
+		} else if (!strcasecmp(name, "restart_processes")) {
+			r = parse_int(value, &config->restart_processes);
+		} else if (!strcasecmp(name, "use_vclocks")) {
+			r = parse_int(value, &config->use_vclocks);
 		} else {
 			pr_err("unknown timemaster setting %s", name);
 			return 1;
@@ -396,6 +418,23 @@ static int parse_timemaster_settings(char **settings,
 	return 0;
 }
 
+static char **parse_raw_settings(char **settings)
+{
+	char **setting, *s, **parsed_settings;
+
+	parsed_settings = (char **)parray_new();
+
+	for (setting = settings; *setting; setting++) {
+		s = *setting;
+		/* Unescape lines beginning with '>' */
+		if (s[0] == '>')
+			s++;
+		parray_append((void ***)&parsed_settings, xstrdup(s));
+	}
+
+	return parsed_settings;
+}
+
 static int parse_section(char **settings, char *name,
 			 struct timemaster_config *config)
 {
@@ -442,8 +481,7 @@ static int parse_section(char **settings, char *name,
 
 	if (settings_dst) {
 		free_parray((void **)*settings_dst);
-		*settings_dst = (char **)parray_new();
-		extend_string_array(settings_dst, settings);
+		*settings_dst = parse_raw_settings(settings);
 	}
 
 	return 0;
@@ -494,6 +532,20 @@ static void config_destroy(struct timemaster_config *config)
 	free(config);
 }
 
+static int check_kernel_version(int version, int patch)
+{
+	struct utsname uts;
+	int v, p;
+
+	if (uname(&uts) < 0)
+		return 1;
+	if (sscanf(uts.release, "%d.%d", &v, &p) < 2)
+		return 1;
+	if (version > v || (version == v && patch > p))
+		return 1;
+	return 0;
+}
+
 static struct timemaster_config *config_parse(char *path)
 {
 	struct timemaster_config *config = xcalloc(1, sizeof(*config));
@@ -506,6 +558,8 @@ static struct timemaster_config *config_parse(char *path)
 	config->ntp_program = DEFAULT_NTP_PROGRAM;
 	config->rundir = xstrdup(DEFAULT_RUNDIR);
 	config->first_shm_segment = DEFAULT_FIRST_SHM_SEGMENT;
+	config->restart_processes = DEFAULT_RESTART_PROCESSES;
+	config->use_vclocks = DEFAULT_USE_VCLOCKS;
 
 	init_program_config(&config->chronyd, "chronyd",
 			    NULL, DEFAULT_CHRONYD_SETTINGS, NULL);
@@ -566,6 +620,9 @@ static struct timemaster_config *config_parse(char *path)
 
 	fclose(f);
 
+	if (config->use_vclocks < 0)
+		config->use_vclocks = !check_kernel_version(5, 18);
+
 	if (section_name)
 		free(section_name);
 	if (section_lines)
@@ -581,7 +638,7 @@ static struct timemaster_config *config_parse(char *path)
 
 static char **get_ptp4l_command(struct program_config *config,
 				struct config_file *file, char **interfaces,
-				int hw_ts)
+				char *phc_index, int hw_ts)
 {
 	char **command = (char **)parray_new();
 
@@ -590,6 +647,9 @@ static char **get_ptp4l_command(struct program_config *config,
 	parray_extend((void ***)&command,
 		      xstrdup("-f"), xstrdup(file->path),
 		      xstrdup(hw_ts ? "-H" : "-S"), NULL);
+	if (phc_index && phc_index[0])
+		parray_extend((void ***)&command,
+			      xstrdup("--phc_index"), xstrdup(phc_index), NULL);
 
 	for (; *interfaces; interfaces++)
 		parray_extend((void ***)&command,
@@ -598,8 +658,9 @@ static char **get_ptp4l_command(struct program_config *config,
 	return command;
 }
 
-static char **get_phc2sys_command(struct program_config *config, int domain,
-				  int poll, int shm_segment, char *uds_path,
+static char **get_phc2sys_command(struct program_config *config,
+				  struct timemaster_config *tconfig, int domain,
+				  int poll, int refclock_id, char *uds_path,
 				  char *message_tag)
 {
 	char **command = (char **)parray_new();
@@ -613,8 +674,25 @@ static char **get_phc2sys_command(struct program_config *config, int domain,
 		      xstrdup("-z"), xstrdup(uds_path),
 		      xstrdup("-t"), xstrdup(message_tag),
 		      xstrdup("-n"), string_newf("%d", domain),
-		      xstrdup("-E"), xstrdup("ntpshm"),
-		      xstrdup("-M"), string_newf("%d", shm_segment), NULL);
+		      xstrdup("-E"), NULL);
+
+	switch (tconfig->ntp_program) {
+	case CHRONYD:
+		parray_extend((void ***)&command,
+			      xstrdup("refclock_sock"),
+			      xstrdup("--refclock_sock_address"),
+			      string_newf("%s/chrony.SOCK%d",
+					  tconfig->rundir, refclock_id),
+			      NULL);
+		break;
+	case NTPD:
+		parray_extend((void ***)&command,
+			      xstrdup("ntpshm"), xstrdup("-M"),
+			      string_newf("%d", refclock_id +
+					  tconfig->first_shm_segment),
+			      NULL);
+		break;
+	}
 
 	return command;
 }
@@ -630,19 +708,36 @@ static char *get_refid(char *prefix, unsigned int number)
 	return NULL;
 };
 
-static void add_shm_source(int shm_segment, int poll, int dpoll, double delay,
-			   char *ntp_options, char *prefix,
-			   struct timemaster_config *config, char **ntp_config)
+static void add_command(char **command, int command_group,
+			struct script *script)
 {
-	char *refid = get_refid(prefix, shm_segment);
+	int *group;
+
+	parray_append((void ***)&script->commands, command);
+
+	group = xmalloc(sizeof(int));
+	*group = command_group;
+	parray_append((void ***)&script->command_groups, group);
+}
+
+static void add_refclock_source(int refclock_id, int poll, int phc_poll,
+				double delay, char *ntp_options, char *prefix,
+				struct timemaster_config *config,
+				char **ntp_config)
+{
+	int filter, shm_segment = refclock_id + config->first_shm_segment;
+	char *refid = get_refid(prefix, refclock_id);
 
 	switch (config->ntp_program) {
 	case CHRONYD:
+		/* set filter to the expected number of samples per poll */
+		filter = (poll >= phc_poll) ? 1 << (poll - phc_poll) : 1;
 		string_appendf(ntp_config,
-			       "refclock SHM %d poll %d dpoll %d "
-			       "refid %s precision 1.0e-9 delay %.1e %s\n",
-			       shm_segment, poll, dpoll, refid, delay,
-			       ntp_options);
+			       "refclock SOCK %s/chrony.SOCK%d poll %d "
+			       "filter %d refid %s precision 1.0e-9 "
+			       "delay %.1e %s\n",
+			       config->rundir, refclock_id, poll,
+			       filter, refid, delay, ntp_options);
 		break;
 	case NTPD:
 		string_appendf(ntp_config,
@@ -667,13 +762,32 @@ static int add_ntp_source(struct ntp_server *source, char **ntp_config)
 	return 0;
 }
 
+static int add_vclock(struct script *script, int pclock_index)
+{
+	struct phc_vclocks **vclocks, *v;
+
+	for (vclocks = script->vclocks; *vclocks; vclocks++) {
+		if ((*vclocks)->pclock_index != pclock_index)
+			continue;
+		return (*vclocks)->vclocks++;
+	}
+
+	v = xmalloc(sizeof(*v));
+	v->pclock_index = pclock_index;
+	v->vclocks = 1;
+	parray_append((void ***)&script->vclocks, v);
+
+	return 0;
+}
+
 static int add_ptp_source(struct ptp_domain *source,
-			  struct timemaster_config *config, int *shm_segment,
-			  int ***allocated_phcs, char **ntp_config,
-			  struct script *script)
+			  struct timemaster_config *config, int *refclock_id,
+			  int *command_group, int ***allocated_phcs,
+			  char **ntp_config, struct script *script)
 {
 	struct config_file *config_file;
-	char **command, *uds_path, **interfaces, *message_tag;
+	char **command, *uds_path, *uds_path2, **interfaces, *message_tag;
+	char ts_interface[IF_NAMESIZE], vclock_index[20];
 	int i, j, num_interfaces, *phc, *phcs, hw_ts, sw_ts;
 	struct sk_ts_info ts_info;
 
@@ -696,26 +810,38 @@ static int add_ptp_source(struct ptp_domain *source,
 	for (i = 0; i < num_interfaces; i++) {
 		phcs[i] = -1;
 
+		/*
+		 * if it is a bonded interface, use the name of the active
+		 * slave interface (which will be timestamping packets)
+		 */
+		if (!rtnl_get_ts_device(source->interfaces[i], ts_interface)) {
+			pr_debug("slave interface of %s: %s",
+				 source->interfaces[i], ts_interface);
+		} else {
+			snprintf(ts_interface, sizeof(ts_interface), "%s",
+				 source->interfaces[i]);
+		}
+
 		/* check if the interface has a usable PHC */
-		if (sk_get_ts_info(source->interfaces[i], &ts_info)) {
+		if (sk_get_ts_info(ts_interface, &ts_info)) {
 			pr_err("failed to get time stamping info for %s",
-			       source->interfaces[i]);
+			       ts_interface);
 			free(phcs);
 			return 1;
 		}
 
 		if (((ts_info.so_timestamping & hw_ts) != hw_ts)) {
-			pr_debug("interface %s: no PHC", source->interfaces[i]);
+			pr_debug("interface %s: no PHC", ts_interface);
 			if ((ts_info.so_timestamping & sw_ts) != sw_ts) {
 				pr_err("time stamping not supported on %s",
-				       source->interfaces[i]);
+				       ts_interface);
 				free(phcs);
 				return 1;
 			}
 			continue;
 		}
 
-		pr_debug("interface %s: PHC %d", source->interfaces[i],
+		pr_debug("interface %s: PHC %d", ts_interface,
 			 ts_info.phc_index);
 
 		/* and the PHC isn't already used in another source */
@@ -749,14 +875,24 @@ static int add_ptp_source(struct ptp_domain *source,
 				}
 			}
 
-			/* don't use this PHC in other sources */
-			phc = xmalloc(sizeof(int));
-			*phc = phcs[i];
-			parray_append((void ***)allocated_phcs, phc);
+			if (config->use_vclocks) {
+				/* request new vclock for the PHC */
+				int vclock = add_vclock(script, phcs[i]);
+				snprintf(vclock_index, sizeof(vclock_index),
+					 "%%PHC%d-%d%%", phcs[i], vclock);
+			} else {
+				/* don't use this PHC in other sources */
+				phc = xmalloc(sizeof(int));
+				*phc = phcs[i];
+				parray_append((void ***)allocated_phcs, phc);
+				vclock_index[0] = '\0';
+			}
 		}
 
 		uds_path = string_newf("%s/ptp4l.%d.socket",
-				       config->rundir, *shm_segment);
+				       config->rundir, *refclock_id);
+		uds_path2 = string_newf("%s/ptp4lro.%d.socket",
+					config->rundir, *refclock_id);
 
 		message_tag = string_newf("[%d", source->domain);
 		for (j = 0; interfaces[j]; j++)
@@ -766,52 +902,74 @@ static int add_ptp_source(struct ptp_domain *source,
 
 		config_file = xmalloc(sizeof(*config_file));
 		config_file->path = string_newf("%s/ptp4l.%d.conf",
-						config->rundir, *shm_segment);
+						config->rundir, *refclock_id);
+
 		config_file->content = xstrdup("[global]\n");
-		extend_config_string(&config_file->content,
-				     config->ptp4l.settings);
+		if (*config->ptp4l.settings) {
+			extend_config_string(&config_file->content,
+					     config->ptp4l.settings);
+			string_appendf(&config_file->content, "\n[global]\n");
+		}
 		extend_config_string(&config_file->content,
 				     source->ptp4l_settings);
 		string_appendf(&config_file->content,
-			       "slaveOnly 1\n"
+			       "clientOnly 1\n"
 			       "domainNumber %d\n"
 			       "uds_address %s\n"
+			       "uds_ro_address %s\n"
 			       "message_tag %s\n",
-			       source->domain, uds_path, message_tag);
+			       source->domain, uds_path, uds_path2,
+			       message_tag);
 
 		if (phcs[i] >= 0) {
 			/* HW time stamping */
 			command = get_ptp4l_command(&config->ptp4l, config_file,
-						    interfaces, 1);
-			parray_append((void ***)&script->commands, command);
+						    interfaces,
+						    vclock_index, 1);
+			add_command(command, *command_group, script);
 
-			command = get_phc2sys_command(&config->phc2sys,
+			command = get_phc2sys_command(&config->phc2sys, config,
 						      source->domain,
 						      source->phc2sys_poll,
-						      *shm_segment, uds_path,
+						      *refclock_id, uds_path,
 						      message_tag);
-			parray_append((void ***)&script->commands, command);
+			add_command(command, (*command_group)++, script);
 		} else {
 			/* SW time stamping */
 			command = get_ptp4l_command(&config->ptp4l, config_file,
-						    interfaces, 0);
-			parray_append((void ***)&script->commands, command);
+						    interfaces, NULL, 0);
+			add_command(command, (*command_group)++, script);
 
-			string_appendf(&config_file->content,
-				       "clock_servo ntpshm\n"
-				       "ntpshm_segment %d\n", *shm_segment);
+			switch (config->ntp_program) {
+			case CHRONYD:
+				string_appendf(&config_file->content,
+					       "clock_servo refclock_sock\n"
+					       "refclock_sock_address "
+					       "%s/chrony.SOCK%d\n",
+					       config->rundir, *refclock_id);
+				break;
+			case NTPD:
+				string_appendf(&config_file->content,
+					       "clock_servo ntpshm\n"
+					       "ntpshm_segment %d\n",
+					       *refclock_id +
+					       config->first_shm_segment);
+				break;
+			}
 		}
 
 		parray_append((void ***)&script->configs, config_file);
 
-		add_shm_source(*shm_segment, source->ntp_poll,
-			       source->phc2sys_poll, source->delay,
-			       source->ntp_options, "PTP", config, ntp_config);
+		add_refclock_source(*refclock_id, source->ntp_poll,
+				    source->phc2sys_poll, source->delay,
+				    source->ntp_options, "PTP", config,
+				    ntp_config);
 
-		(*shm_segment)++;
+		(*refclock_id)++;
 
 		free(message_tag);
 		free(uds_path);
+		free(uds_path2);
 		free(interfaces);
 	}
 
@@ -847,7 +1005,8 @@ static char **get_ntpd_command(struct program_config *config,
 }
 
 static struct config_file *add_ntp_program(struct timemaster_config *config,
-					   struct script *script)
+					   struct script *script,
+					   int command_group)
 {
 	struct config_file *ntp_config = xmalloc(sizeof(*ntp_config));
 	char **command = NULL;
@@ -871,7 +1030,7 @@ static struct config_file *add_ntp_program(struct timemaster_config *config,
 	}
 
 	parray_append((void ***)&script->configs, ntp_config);
-	parray_append((void ***)&script->commands, command);
+	add_command(command, command_group, script);
 
 	return ntp_config;
 }
@@ -879,7 +1038,13 @@ static struct config_file *add_ntp_program(struct timemaster_config *config,
 static void script_destroy(struct script *script)
 {
 	char ***commands, **command;
+	int **groups;
 	struct config_file *config, **configs;
+	struct phc_vclocks **vclocks;
+
+	for (vclocks = script->vclocks; *vclocks; vclocks++)
+		free(*vclocks);
+	free(script->vclocks);
 
 	for (configs = script->configs; *configs; configs++) {
 		config = *configs;
@@ -896,6 +1061,10 @@ static void script_destroy(struct script *script)
 	}
 	free(script->commands);
 
+	for (groups = script->command_groups; *groups; groups++)
+		free(*groups);
+	free(script->command_groups);
+
 	free(script);
 }
 
@@ -905,13 +1074,16 @@ static struct script *script_create(struct timemaster_config *config)
 	struct source *source, **sources;
 	struct config_file *ntp_config = NULL;
 	int **allocated_phcs = (int **)parray_new();
-	int ret = 0, shm_segment;
+	int ret = 0, refclock_id = 0, command_group = 0;
 
+	script->vclocks = (struct phc_vclocks **)parray_new();
 	script->configs = (struct config_file **)parray_new();
 	script->commands = (char ***)parray_new();
+	script->command_groups = (int **)parray_new();
+	script->no_restart_group = command_group;
+	script->restart_groups = config->restart_processes;
 
-	ntp_config = add_ntp_program(config, script);
-	shm_segment = config->first_shm_segment;
+	ntp_config = add_ntp_program(config, script, command_group++);
 
 	for (sources = config->sources; (source = *sources); sources++) {
 		switch (source->type) {
@@ -920,8 +1092,8 @@ static struct script *script_create(struct timemaster_config *config)
 				ret = 1;
 			break;
 		case PTP_DOMAIN:
-			if (add_ptp_source(&source->ptp, config, &shm_segment,
-					   &allocated_phcs,
+			if (add_ptp_source(&source->ptp, config, &refclock_id,
+					   &command_group, &allocated_phcs,
 					   &ntp_config->content, script))
 				ret = 1;
 			break;
@@ -1046,12 +1218,109 @@ static int remove_config_files(struct config_file **configs)
 	return 0;
 }
 
+static int set_phc_n_vclocks(int phc_index, int n_vclocks)
+{
+	char path[PATH_MAX];
+	FILE *f;
+
+	snprintf(path, sizeof(path), "/sys/class/ptp/ptp%d/n_vclocks",
+		 phc_index);
+	f = fopen(path, "w");
+	if (!f) {
+		pr_err("failed to open %s: %m", path);
+		return 1;
+	}
+	fprintf(f, "%d\n", n_vclocks);
+	fclose(f);
+
+	return 0;
+}
+
+static int create_vclocks(struct phc_vclocks **phc_vclocks)
+{
+	struct phc_vclocks **vclocks;
+
+	for (vclocks = phc_vclocks; *vclocks; vclocks++) {
+		if (set_phc_n_vclocks((*vclocks)->pclock_index,
+				      (*vclocks)->vclocks))
+			return 1;
+	}
+
+	return 0;
+}
+
+static int remove_vclocks(struct phc_vclocks **phc_vclocks)
+{
+	struct phc_vclocks **vclocks;
+
+	for (vclocks = phc_vclocks; *vclocks; vclocks++) {
+		if (set_phc_n_vclocks((*vclocks)->pclock_index, 0))
+			return 1;
+	}
+
+	return 0;
+}
+
+static int get_vclock_index(int pindex, int vclock)
+{
+	char pattern[PATH_MAX], *s;
+	int n, vindex;
+	glob_t gl;
+
+	snprintf(pattern, sizeof(pattern), "/sys/class/ptp/ptp%d/ptp[0-9]*",
+		 pindex);
+
+	if (glob(pattern, 0, NULL, &gl)) {
+		pr_err("glob(%s) failed", pattern);
+		return -1;
+	}
+
+	if (vclock >= gl.gl_pathc ||
+	    !(s = strrchr(gl.gl_pathv[vclock], '/')) ||
+	    sscanf(s + 1, "ptp%d%n", &vindex, &n) != 1 ||
+	    n != strlen(s + 1)) {
+		pr_err("missing vclock %d:%d", pindex, vclock);
+		globfree(&gl);
+		return -1;
+	}
+
+	globfree(&gl);
+
+	return vindex;
+}
+
+static int translate_vclock_options(char ***commands)
+{
+	int n, pindex, vclock, vindex, blen;
+	char **command;
+
+	for (; *commands; commands++) {
+		for (command = *commands; *command; command++) {
+			if (sscanf(*command, "%%PHC%d-%d%%%n",
+				   &pindex, &vclock, &n) != 2 ||
+			    n != strlen(*command))
+				continue;
+			vindex = get_vclock_index(pindex, vclock);
+			if (vindex < 0)
+				return 1;
+
+			/* overwrite the string with the vclock PHC index */
+			blen = strlen(*command) + 1;
+			if (snprintf(*command, blen, "%d", vindex) >= blen)
+				return 1;
+		}
+	}
+
+	return 0;
+}
+
 static int script_run(struct script *script)
 {
+	struct timespec ts_start, ts_now;
 	sigset_t mask, old_mask;
 	siginfo_t info;
 	pid_t pid, *pids;
-	int i, num_commands, status, ret = 0;
+	int i, group, num_commands, status, quit = 0, ret = 0;
 
 	for (num_commands = 0; script->commands[num_commands]; num_commands++)
 		;
@@ -1064,6 +1333,12 @@ static int script_run(struct script *script)
 	if (create_config_files(script->configs))
 		return 1;
 
+	if (create_vclocks(script->vclocks))
+		return 1;
+
+	if (translate_vclock_options(script->commands))
+		return 1;
+
 	sigemptyset(&mask);
 	sigaddset(&mask, SIGCHLD);
 	sigaddset(&mask, SIGTERM);
@@ -1086,7 +1361,9 @@ static int script_run(struct script *script)
 		}
 	}
 
-	/* wait for one of the blocked signals */
+	clock_gettime(CLOCK_MONOTONIC, &ts_start);
+
+	/* process the blocked signals */
 	while (1) {
 		if (sigwaitinfo(&mask, &info) < 0) {
 			if (errno == EINTR)
@@ -1095,41 +1372,119 @@ static int script_run(struct script *script)
 			break;
 		}
 
-		/*
-		 * assume only the first process (i.e. chronyd or ntpd) is
-		 * essential and continue if other processes terminate
-		 */
-		if (info.si_signo == SIGCHLD && info.si_pid != pids[0]) {
-			pr_info("process %d terminated (ignored)", info.si_pid);
+		clock_gettime(CLOCK_MONOTONIC, &ts_now);
+
+		if (info.si_signo != SIGCHLD) {
+			if (quit)
+				continue;
+
+			quit = 1;
+			pr_debug("exiting on signal %d", info.si_signo);
+
+			/* terminate remaining processes */
+			for (i = 0; i < num_commands; i++) {
+				if (pids[i] > 0) {
+					pr_debug("killing process %d", pids[i]);
+					kill(pids[i], SIGTERM);
+				}
+			}
+
 			continue;
 		}
 
-		pr_info("received signal %d", info.si_signo);
-		break;
-	}
+		/* wait for all terminated processes */
+		while (1) {
+			pid = waitpid(-1, &status, WNOHANG);
+			if (pid <= 0)
+				break;
 
-	/* kill all started processes */
-	for (i = 0; i < num_commands; i++) {
-		if (pids[i] > 0) {
-			pr_debug("killing process %d", pids[i]);
-			kill(pids[i], SIGTERM);
+			if (!WIFEXITED(status)) {
+				pr_info("process %d terminated abnormally",
+					pid);
+			} else {
+				pr_info("process %d terminated with status %d",
+					pid, WEXITSTATUS(status));
+			}
+
+			for (i = 0; i < num_commands; i++) {
+				if (pids[i] == pid)
+					pids[i] = 0;
+			}
 		}
-	}
 
-	while ((pid = wait(&status)) >= 0) {
-		if (!WIFEXITED(status)) {
-			pr_info("process %d terminated abnormally", pid);
-			ret = 1;
-		} else {
-			if (WEXITSTATUS(status))
+		/* wait for all processes to terminate when exiting */
+		if (quit) {
+			for (i = 0; i < num_commands; i++) {
+				if (pids[i])
+					break;
+			}
+			if (i == num_commands)
+				break;
+
+			pr_debug("waiting for other processes to terminate");
+			continue;
+		}
+
+		/*
+		 * terminate (and then restart if allowed) all processes in
+		 * groups that have a terminated process
+		 */
+		for (group = 0; group < num_commands; group++) {
+			int terminated = 0, running = 0;
+
+			for (i = 0; i < num_commands; i++) {
+				if (*(script->command_groups[i]) != group)
+					continue;
+				if (pids[i])
+					running++;
+				else
+					terminated++;
+			}
+
+			if (!terminated)
+				continue;
+
+			/*
+			 * exit with a non-zero status if the group should not
+			 * be restarted (i.e. chronyd/ntpd), timemaster is
+			 * running only for a short time (and it is likely a
+			 * configuration error), or restarting is disabled
+			 * completely
+			 */
+			if (group == script->no_restart_group ||
+			    ts_now.tv_sec - ts_start.tv_sec <= 1 ||
+			    !script->restart_groups) {
+				kill(getpid(), SIGTERM);
 				ret = 1;
-			pr_info("process %d terminated with status %d", pid,
-				WEXITSTATUS(status));
+				break;
+			}
+
+			for (i = 0; i < num_commands; i++) {
+				if (*(script->command_groups[i]) != group)
+					continue;
+
+				/* terminate all processes in the group first */
+				if (running && pids[i]) {
+					pr_debug("killing process %d", pids[i]);
+					kill(pids[i], SIGTERM);
+				} else if (!running && !pids[i]) {
+					pids[i] = start_program(script->commands[i],
+								&old_mask);
+					if (!pids[i])
+						kill(getpid(), SIGTERM);
+
+					/* limit restarting rate */
+					sleep(1);
+				}
+			}
 		}
 	}
 
 	free(pids);
 
+	if (remove_vclocks(script->vclocks))
+		return 1;
+
 	if (remove_config_files(script->configs))
 		return 1;
 
@@ -1139,15 +1494,25 @@ static int script_run(struct script *script)
 static void script_print(struct script *script)
 {
 	char ***commands, **command;
+	int **groups;
 	struct config_file *config, **configs;
+	struct phc_vclocks **vclocks;
 
 	for (configs = script->configs; *configs; configs++) {
 		config = *configs;
 		fprintf(stderr, "%s:\n\n%s\n", config->path, config->content);
 	}
 
-	fprintf(stderr, "commands:\n\n");
-	for (commands = script->commands; *commands; commands++) {
+	fprintf(stderr, "virtual clocks:\n\n");
+	for (vclocks = script->vclocks; *vclocks; vclocks++) {
+		fprintf(stderr, "PHC%d: %d\n",
+			(*vclocks)->pclock_index, (*vclocks)->vclocks);
+	}
+
+	fprintf(stderr, "\ncommands:\n\n");
+	for (commands = script->commands, groups = script->command_groups;
+	     *commands; commands++, groups++) {
+		fprintf(stderr, "[%d] ", **groups);
 		for (command = *commands; *command; command++)
 			fprintf(stderr, "%s ", *command);
 		fprintf(stderr, "\n");
