@@ -18,6 +18,7 @@
  */
 #include <arpa/inet.h>
 #include <errno.h>
+#include <linux/limits.h>
 #include <signal.h>
 #include <stdarg.h>
 #include <stdio.h>
@@ -190,6 +191,35 @@ char *portaddr2str(struct PortAddress *addr)
 	return buf;
 }
 
+const char *ustate2str(enum unicast_state ustate)
+{
+	switch (ustate) {
+	case UC_WAIT:
+		return "WAIT";
+	case UC_HAVE_ANN:
+		return "HAVE_ANN";
+	case UC_NEED_SYDY:
+		return "NEED_SYDY";
+	case UC_HAVE_SYDY:
+		return "HAVE_SYDY";
+	}
+
+	return "???";
+}
+
+enum port_state port_state_normalize(enum port_state state)
+{
+	switch (state) {
+	case PS_MASTER:
+	case PS_SLAVE:
+	case PS_PRE_MASTER:
+	case PS_UNCALIBRATED:
+		return state;
+	default:
+		return PS_DISABLED;
+	}
+}
+
 void posix_clock_close(clockid_t clock)
 {
 	if (clock == CLOCK_REALTIME) {
@@ -200,6 +230,7 @@ void posix_clock_close(clockid_t clock)
 
 clockid_t posix_clock_open(const char *device, int *phc_index)
 {
+	char phc_device_path[PATH_MAX];
 	struct sk_ts_info ts_info;
 	char phc_device[19];
 	int clkid;
@@ -208,21 +239,29 @@ clockid_t posix_clock_open(const char *device, int *phc_index)
 	if (!strcasecmp(device, "CLOCK_REALTIME")) {
 		return CLOCK_REALTIME;
 	}
-	/* check if device is valid phc device */
-	clkid = phc_open(device);
-	if (clkid != CLOCK_INVALID) {
-		if (!strncmp(device, "/dev/ptp", strlen("/dev/ptp"))) {
-			int r = get_ranged_int(device + strlen("/dev/ptp"),
+
+	/* if the device name resolves so a plausible filesystem path, we
+	 * assume it is the path to a PHC char device, and treat it as such
+	 */
+	if (realpath(device, phc_device_path)) {
+		clkid = phc_open(device);
+		if (clkid == CLOCK_INVALID)
+			return clkid;
+
+		if (!strncmp(phc_device_path, "/dev/ptp", strlen("/dev/ptp"))) {
+			int r = get_ranged_int(phc_device_path + strlen("/dev/ptp"),
 					       phc_index, 0, 65535);
 			if (r) {
 				fprintf(stderr,
 					"failed to parse PHC index from %s\n",
-					device);
-				return -1;
+					phc_device_path);
+				phc_close(clkid);
+				return CLOCK_INVALID;
 			}
 		}
 		return clkid;
 	}
+
 	/* check if device is a valid ethernet device */
 	if (sk_get_ts_info(device, &ts_info) || !ts_info.valid) {
 		pr_err("unknown clock %s: %m", device);
@@ -582,6 +621,10 @@ int handle_term_signals(void)
 		fprintf(stderr, "cannot handle SIGTERM\n");
 		return -1;
 	}
+	if (SIG_ERR == signal(SIGHUP, handle_int_quit_term)) {
+		fprintf(stderr, "cannot handle SIGHUP\n");
+		return -1;
+	}
 	return 0;
 }
 
